# -*- coding: ISO-8859-1 -*-

############################################################################## 
## 2004, Arnulf Heimsbakk
## 
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
## IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
## ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR 
## ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
## DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
## SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
## CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
## LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
## OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
## SUCH DAMAGE.
##
## Otherwise the GPL license: http://www.gnu.org/licenses/gpl.txt
##############################################################################

import OFS.SimpleItem, OFS.ObjectManager, Acquisition, AccessControl.Role
import Globals, OFS.PropertyManager, OFS.CopySupport, OFS.Image, webdav.Collection
import OFS.Folder
from AccessControl import ClassSecurityInfo
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from ZPublisher.Iterators import filestream_iterator
from Acquisition import aq_base
from TreeDisplay import TreeTag
from DateTime import DateTime
from OFS.CopySupport import _cb_decode, cookie_path, sanity_check
from OFS import Moniker
from cgi import escape
import webdav 
from zExceptions import NotFound

import __builtin__
import os, string, framer, stat, effects, types, traceback, EXIF
from PIL import Image
from thumbnailer import thumbnailer
from md5 import md5

_marker = []

def _get_cache_dirs():
    path = os.path.join(__builtin__.CLIENT_HOME, "lis")
    return os.path.join(path, "cache"), os.path.join(path, "meta")

def _aq_first(self, ancestor, name, object, extra):
    return isinstance(object, LocalImageSystem)

class LocalFolder(OFS.PropertyManager.PropertyManager,
                  Globals.Persistent,
                  OFS.ObjectManager.ObjectManager,
                  Acquisition.Implicit,
                  AccessControl.Role.RoleManager,
                  OFS.SimpleItem.Item):
    """Represent a local folder as a Zope Folder-like object. It search the 
    local directory for images and represents them as Image-like boject. 
    This folder hides doted fiels and directories which are hidden files in 
    Linux. """
    
    manage_options = (
        {'label':'Contents', 'action':'manage_main'}, ) + \
        OFS.PropertyManager.PropertyManager.manage_options + \
        AccessControl.Role.RoleManager.manage_options

    meta_type = "Local Folder"
    icon = 'misc_/LocalImageSystem/Folder_icon.gif'
    
    all_meta_types = (
        {'name': 'Local Image', 'action': 'manage_addImageForm', 'permission': 'Add local image'},
        {'name': 'Local File', 'action': 'manage_addFileForm', 'permission': 'Add local file'},
        {'name': 'Local Folder', 'action': 'manage_addFolderForm','permission': 'Add local folder'},
        )  # should only contain IFS items
                       
    #~ filtered_meta_types = None
    
    isPrincipiaFolderish = 1
    security = ClassSecurityInfo()
    security.declareObjectProtected("View")    
    
    security.declarePrivate('dummy')
    dummy = None
    
    def __init__(self, id=None, path=None):
        if id != None: self.id = id 
        if path != None: self.path = path
        self._uid = ''

    def _subobject_permissions(self):
        return self.aq_acquire('_getProductRegistryData')('ac_permissions')

    def __bobo_traverse__(self, REQUEST, id):
        try: return self._getOb(id)
        except: pass
        
        method = REQUEST.get('REQUEST_METHOD', 'GET')
        if method not in ('GET', 'POST'):
            return LocalFile(id, '').__of__(self)
                
    def _getOb(self, id, default=_marker):
        fullpath = os.path.join(self.path, id)
        file_or_dir = False
        ob = None
        
        try:
            if os.path.exists(fullpath):
                if os.path.isdir(fullpath):
                    file_or_dir = True
                    uid, uname = self._uname(fullpath, 'lis_folder_')
                    if not hasattr(aq_base(self), uname):
                        self._setObject(uname, LocalFolder(id, fullpath), set_owner=0)
                    id = uname
                elif os.path.isfile(fullpath):
                    file_or_dir = True
                    uid, uname = self._uname(fullpath, 'lis_image_')
                    if not hasattr(aq_base(self), uname):
                        meta = self._get_image_meta(fullpath, uname, uid, id)
                        if meta == None: 
                            ob = LocalFile(id, fullpath)
                            return ob.__of__(self)
                    id = uname
        except:
            pass

        if id[:1] != '_' and hasattr(aq_base(self), id):
            ob = getattr(self, id)
            if file_or_dir:
                if ob == []: return ob
                if ob.path != fullpath:
                    ob.path = fullpath
                if ob._uid != uid:
                    ob._uid = uid
            return ob
        if default is _marker:
            raise AttributeError, id
        return default
        
    def _delOb(self, id, dp=1):
        fullpath = os.path.join(self.path, id)
        
        if os.path.exists(fullpath):
            if os.path.isdir(fullpath):
                uid, id = self._uname(fullpath, 'lis_folder_')
                for root, dirs, files in os.walk(fullpath, topdown=False):
                    for name in files: os.remove(os.path.join(root, name))
                    for name in dirs: os.rmdir(os.path.join(root, name))
                os.rmdir(fullpath)
            elif os.path.isfile(fullpath):
                uid, id = self._uname(fullpath, 'lis_image_')
                os.remove(fullpath)
        
        try: OFS.ObjectManager.ObjectManager._delOb(self, id)
        except: pass
        
    def tpId(self): 
        ob = self
        ids = []
       
        while True:
            if not hasattr(ob, 'id') or type(ob.id) != type(""): break
            ids.append(ob.id)
            if not hasattr(ob, 'aq_parent'): break
            ob = ob.aq_parent
        
        ids.reverse()

        return TreeTag.encode_str(string.join(ids, '/'))

                                                             
    security.declarePublic('index_html')
    def index_html(self, spec=None):
        """Simple thumbnail view of folder."""
        
        return ""
        
    security.declareProtected('View management screens', 'manage_main')
    manage_main=Globals.DTMLFile('dtml/main', globals())    

    security.declareProtected('Add local folder', 'manage_addFolder',
                                                  'manage_addFolderForm')

    manage_addFolderForm=PageTemplateFile('zpt/addFolder.pt', globals())    
    def manage_addFolder(self, name, REQUEST=None):
        """Create a folder on the local file system."""
        name = name.strip()
        
        if name != '':
            new_folder = os.path.join(self.path, name)
            os.mkdir(new_folder)
         
        if REQUEST != None:
            return self.manage_main(self, REQUEST, 
                                    management_view="Content",
                                    manage_tabs_message="Local Folder added.")


    security.declareProtected('Add local image', 'manage_addImage',
                                                 'manage_addImageForm')
    manage_addImageForm=PageTemplateFile('zpt/addImage.pt', globals())    
    def manage_addImage(self, image, id='', REQUEST=None):
        """Upload a image to the local filesystem."""

        if hasattr(image,'filename'):
            fullname=image.filename
        elif hasattr(image,'name'):
            fullname=image.name
        else:
            fullname = id
            
        if id == '': id = os.path.split(fullname)[-1]
        dest_name = os.path.join(self.path, id)
        file(dest_name, "wb").write(image.read())
         
        if REQUEST != None:
            return self.manage_main(self, REQUEST, 
                                    management_view="Content",
                                    manage_tabs_message="Local Image added.")

    security.declareProtected("Add local file", 'manage_addFileForm')
    manage_addFileForm=PageTemplateFile('zpt/addFile.pt', globals())

    security.declareProtected('Delete objects', 'manage_delObjects')
    security.declareProtected('Copy or Move', 'manage_cutObjects',
                                              'manage_pasteObjects',
                                              'manage_copyObjects') 
    security.declareProtected('Manage properties', 'manage_renameObject',
                                                   'manage_renameForm',
                                                   'manage_renameObjects')
                                                   
    def manage_renameObject(self, id, new_id, REQUEST=None):
        """Rename a image or a folder."""
        
        fullpath = os.path.join(self.path, id)
        new_fullpath = os.path.join(self.path, new_id)
        
        ob = self._getOb(id)
        if ob != None and ob.meta_type == 'Local Unsupported Image':
            try: os.rename(fullpath, new_fullpath)
            except: pass
        else:
            meta = ""
            if os.path.isdir(fullpath):     meta = "folder"
            elif os.path.isfile(fullpath):  meta = "image"
            
            if meta == "": return None
    
            try:
                uid, uname = self._uname(fullpath, 'lis_%s_' % meta)
                os.rename(fullpath, new_fullpath)
                new_uid, new_uname = self._uname(new_fullpath, 'lis_%s_' % meta)
                ob = getattr(self, uname)
                ob.path = new_fullpath
                ob._uid = new_uid
                ob._setId(new_id)
            except OSError: pass
            except Exception, e: 
                os.rename(new_fullpath, fullpath)
                raise e

        if REQUEST is not None:
            return self.manage_main(self, REQUEST, update_menu=1)
        return None

    def manage_copyObjects(self, ids=None, REQUEST=None, RESPONSE=None):
        """Put a reference to the objects named in ids in the clip board"""
        lis_ids = []

        if ids != None:
            if type(ids) is type(''):
                ids=[ids]
    
            for id in ids:
                fullpath = os.path.join(self.path, id)
                meta = ""
                if os.path.isdir(fullpath):     meta = "folder"
                elif os.path.isfile(fullpath):  meta = "image"
                if meta != "": 
                    uid, uname = self._uname(fullpath, 'lis_%s_' % meta)
                    if hasattr(self, uname):
                        lis_ids.append(uname)
                    else:
                        ob = self._getOb(id)
                        if ob != None and ob.meta_type == 'Local Image':
                            lis_ids.append(id)
                
            if len(lis_ids):
                ids = lis_ids        
        
        return OFS.CopySupport.CopyContainer.manage_copyObjects(self, ids, REQUEST, RESPONSE)

    def manage_pasteObjects(self, cb_copy_data=None, REQUEST=None):
        """Paste previously copied objects into the current object.
           If calling manage_pasteObjects from python code, pass
           the result of a previous call to manage_cutObjects or
           manage_copyObjects as the first argument."""

        cp = None        
        if cb_copy_data is not None:
            cp=cb_copy_data
        else:
            if REQUEST and REQUEST.has_key('__cp'):
                cp=REQUEST['__cp']
        if cp is None:
            raise CopyError, eNoData
        
        cp_orig = cp
        try:    cp=_cb_decode(cp)
        except: 
            raise CopyError, eInvalid

        oblist=[]
        op=cp[0]
        app = self.getPhysicalRoot()
        result = []

        for mdata in cp[1]:
            m = Moniker.loadMoniker(mdata)
            try: 
                ob = m.bind(app)
            except: 
                print traceback.print_exc()
                raise CopyError, eNotFound
            self._verifyObjectPaste(ob, validate_src=op+1)
            oblist.append(ob)

        for i in oblist:
            if i.meta_type == "Local Image":
                result.append(self._paste_image(i, self, op))
            elif i.meta_type == 'Local Unsupported Image':
                result.append(self._paste_file(i, self, op))
            elif i.meta_type == "Local Folder":
                result.append(self._paste_dir(i, self, op))
        
        if REQUEST is not None:
            REQUEST['RESPONSE'].setCookie('__cp', 'deleted',
                                path='%s' % cookie_path(REQUEST),
                                expires='Wed, 31-Dec-97 23:59:59 GMT')
            REQUEST['__cp'] = None
            return self.manage_main(self, REQUEST, update_menu=1,
                                    cb_dataValid=0)
                                    
        return result

    def _paste_dir(self, ob, dest, op=0):
        if not ob.cb_isCopyable():
            raise CopyError, eNotSupported % escape(ob.getId())
        try:    ob._notifyOfCopyTo(self, op=0)
        except: raise CopyError, MessageDialog(
            title='Copy Error',
            message=sys.exc_info()[1],
            action ='manage_main')

        if not sanity_check(self, ob):
            raise 'Copy/Move Error', 'This object cannot be pasted into itself'

        ob_path = ob.path

        if op == 1 and ob.aq_parent == dest:
            raise 'Move Error', 'This object cannot be pasted into same folder'
            
        new_id = self._get_id(ob.getId())

        new_ob_path = os.path.join(dest.path, new_id)
        os.makedirs(new_ob_path)
                
        uid, uname = self._uname(new_ob_path, 'lis_folder_')
        new_ob = ob._getCopy(self)
        new_ob.path = new_ob_path
        new_ob._uid = uid
        
        new_ob._setId(new_id)
        objects = ob.objectValues()
        
        dest._setObject(uname, new_ob, set_owner=0)

        for o in objects:
            if o.meta_type == "Local Folder":
                self._paste_dir(o, new_ob, op)

        for o in objects:
            if o.meta_type == "Local Image":
                self._paste_image(o, new_ob, op)
            elif o.meta_type == 'Local Unsupported Image':
                self._paste_file(o, new_ob, op)
    
        if op == 1:
            try: os.rmdir(ob.path)
            except: pass
            
        return {'id': ob.getId(), 'new_id': new_ob.getId()}
            
    def _paste_file(self, ob, dest, op=0):
        if not ob.cb_isCopyable():
            raise CopyError, eNotSupported % escape(ob.getId())
        try:    ob._notifyOfCopyTo(self, op=0)
        except: raise CopyError, MessageDialog(
            title='Copy Error',
            message=sys.exc_info()[1],
            action ='manage_main')

        if op != 0 and ob.aq_parent == dest:
            new_id = ob.getId()
        else:
            new_id = self._get_id(ob.getId())

        file_path = ob.path
        new_file_path = os.path.join(dest.path, new_id)
        if op == 0:
            file(new_file_path, "wb").write(file(ob.path).read())
        else:
            os.rename(ob.path, new_file_path)

    def _paste_image(self, ob, dest, op=0):
        if not ob.cb_isCopyable():
            raise CopyError, eNotSupported % escape(ob.getId())
        try:    ob._notifyOfCopyTo(self, op=0)
        except: raise CopyError, MessageDialog(
            title='Copy Error',
            message=sys.exc_info()[1],
            action ='manage_main')

        if op == 0:
            new_ob = ob._getCopy(self)
        else:
            new_ob = ob
            uid, uname = self._uname(ob.path, 'lis_image_')
            ob.aq_parent._delObject(uname)
        
        id = ob.getId()
        if op != 0 and ob.aq_parent == dest:
            new_id = new_ob.getId()
        else:
            new_id = self._get_id(new_ob.getId())
            
        new_ob._setId(new_id)
        img_path = ob.path
        new_img_path = os.path.join(dest.path, new_id)
        
        if op == 0:
            file(new_img_path, "wb").write(file(ob.path).read())
        else:
            os.rename(ob.path, new_img_path)

        uid, uname = self._uname(new_img_path, 'lis_image_')
        new_ob._uid = uid
        try: dest._setObject(uname, new_ob, set_owner=0)
        except: pass
        
        return {'id': id, 'new_id': new_id}

    security.declareProtected('Access contents information', 'bobobase_modification_time',
                                                             'objectIds')
    def objectIds(self, spec=None):
        all_files = []
        all_dirs = []
        image_files = []
        unames = []
        specs = []
        if spec:
            specs = spec.split(",")
        
        if os.path.exists(self.path):
            for root, dirs, files in os.walk(self.path):
                dirs = [d for d in dirs if d[0] != '.']
                
                for i in dirs:
                    unames.append(self._uname(os.path.join(root, i), 'lis_folder_')[1])
                    if "Local Folder" in specs or spec == None:
                        all_dirs.append(i)
                    
                files = [f for f in files if f[0] != '.']
                for i in files:
                    fullpath = os.path.join(root, i)
                    try:
                        uid, uname = self._uname(fullpath, 'lis_image_')
                    except: continue
                    unames.append(uname)
                    if self._get_image_meta(fullpath, uname, uid, i) == None: 
                        if "Local File" in specs or spec == None:
                            all_files.append(i)
                        continue
                    else:
                        if "Local Image" in specs or spec == None:
                            all_files.append(i)
                        image_files.append(i)
                    
                all_files += all_dirs #+ files #image_files
                break
                
        #~ print [i['id'] for i in aq_base(self)._objects if i['id']  not in unames]
        for f in [i['id'] for i in aq_base(self)._objects if i['id']  not in unames]:
            try:
                aq_base(self)._delObject(f)
            except: pass
                

        return all_files

    def bobobase_modification_time(self):
        try: 
            return DateTime(os.stat(self.path)[stat.ST_MTIME])
        except: 
            return DateTime()

    
    def _uname(self, filename, pre=''):
        try: 
            s = os.stat(filename)
            path, name = os.path.split(filename)
        except: s = None

        if s[stat.ST_INO] == 0: to_md5 = (s[stat.ST_CTIME], ) # ntfs/fat
        else:                   to_md5 = (s[stat.ST_INO], ) # inodes 
            
        if not os.path.isdir(filename):
            to_md5 += (s[stat.ST_SIZE], s[stat.ST_MTIME])
        
        if s != None:
            uname = md5(str(to_md5)).hexdigest()

            return uname, pre + uname
        return None, None

    def _get_image_meta(self, fullpath, uname, uid, id):
        t = thumbnailer(_get_cache_dirs())
        if hasattr(aq_base(self), uname):  
            meta = getattr(self, uname)._meta
        else:
            meta = t.is_image(fullpath)
            if meta == None: return None
            
            self._setObject(uname, LocalImage(id, fullpath, uid, meta), set_owner=0)
        return meta

    def cb_isMoveable(self):
        ret = OFS.CopySupport.CopySource.cb_isMoveable(self)
        return os.access(self.path, os.R_OK | os.W_OK) and ret
        
    def cb_isCopyable(self):
        ret = OFS.CopySupport.CopySource.cb_isCopyable(self)
        return os.access(self.path, os.R_OK) and ret


    def manage_addLocalImage(dispatcher, id, path, uid, meta, REQUEST=None):
        """
        Create a new Local Image in IFS.
        """
        dispatcher.Destination()._setObject(id, LocalImage(id, path, uid, meta))
        if REQUEST != None:
            return dispatcher.manage_main(dispatcher, REQUEST) 

    security.declareProtected("Add local folder", "manage_addFolder", "manage_addLocalFolder")
    def manage_addFolder(dispatcher, id, REQUEST=None):
        """
        Add a folder to the disk (FTP)
        """
        os.mkdir(os.path.join(dispatcher.path, id))
        if REQUEST != None:
            return dispatcher.manage_main(dispatcher, REQUEST) 

    def manage_addLocalFolder(dispatcher, id, path, REQUEST=None):
        """
        Create a new Local Folder in IFS.
        """
        dispatcher.Destination()._setObject(id, LocalImage(id, path, uid, meta))
        if REQUEST != None:
            return dispatcher.manage_main(dispatcher, REQUEST) 



Globals.InitializeClass(LocalFolder)

class LocalFile(OFS.SimpleItem.Item,
                AccessControl.Role.RoleManager,
                Acquisition.Implicit):
    """
    Dummy object for unsupported files.
    """
    
    meta_type = "Local File"
    icon = "misc_/LocalImageSystem/Unknown_icon.gif"
                
    security = ClassSecurityInfo()  
    security.declareObjectProtected("View")    

    manage_options = ({'label':'Unsupported file', 'action':'manage_main'}, )

    def __init__(self, id, path, request=None):
        self.id = id
        self.path = path
        try: self.stat = os.stat(path)
        except: self.stat = None

    def __bobo_traverse__(self, REQUEST, name=None):
        # We must handle traversal so that we can recognize situations
        # where a 409 Conflict must be returned instead of the normal
        # 404 Not Found, per [WebDAV 8.3.1].
        try:    return getattr(self, name)
        except: pass
        method=REQUEST.get('REQUEST_METHOD', 'GET')
        if method in ('PUT', 'MKCOL', 'LOCK'):
            raise Conflict, 'Collection ancestors must already exist.'
        raise NotFound, 'The requested resource was not found.'

    security.declareProtected('Manage properties', 'manage_main')
    manage_main = PageTemplateFile('zpt/unknown.pt', globals())   

    security.declarePublic('HEAD')
    def HEAD(self, REQUEST, RESPONSE):
        """Retrieve resource information without a response message body."""
        self.dav__init(REQUEST, RESPONSE)
        raise NotFound, 'The requested resource does not exist.'

    # Most methods return 404 (Not Found)
    DELETE=TRACE=PROPFIND=PROPPATCH=COPY=MOVE=HEAD

    def cb_isMoveable(self):
        ret = OFS.CopySupport.CopySource.cb_isMoveable(self)
        return os.access(self.path, os.R_OK | os.W_OK) and ret
        
    def cb_isCopyable(self):
        ret = OFS.CopySupport.CopySource.cb_isCopyable(self)
        return os.access(self.path, os.R_OK) and ret

    def bobobase_modification_time(self):
        if self.stat != None:
            return DateTime(self.stat[stat.ST_MTIME])

    def get_size(self):
        if self.stat != None:
            return self.stat[stat.ST_SIZE]

    def manage_FTPget(self):
        """Get file for FTP."""
        return file(self.path, "rb").read()

    security.declareProtected("Add local folder", "MKCOL")
    def MKCOL(self, REQUEST, RESPONSE):
        """Handle HTTP MKCOL request."""
        self.dav__init(REQUEST, RESPONSE)
        if REQUEST.get('BODY', ''):
            raise UnsupportedMediaType, 'Unknown request body.'

        self.aq_parent.manage_addFolder(self.id)

        RESPONSE.setStatus(201)
        RESPONSE.setBody('')
        return RESPONSE

    security.declareProtected('Add local file', 'PUT')
    def PUT(self, REQUEST=None, RESPONSE=None):
        """Handle HTTP PUT requests."""
        self.dav__init(REQUEST, RESPONSE)
        
        if REQUEST.has_key("BODYFILE"):
            try:
                file(os.path.join(self.aq_parent.path, self.id), "wb").write(REQUEST["BODYFILE"].read())
            except: pass
        
        RESPONSE.setStatus(201)
        RESPONSE.setBody('')
        return RESPONSE


Globals.InitializeClass(LocalFile)

class LocalImage(Globals.Persistent,
                 OFS.SimpleItem.Item,
                 AccessControl.Role.RoleManager,
                 Acquisition.Implicit,
                 OFS.PropertyManager.PropertyManager):
    """Represent a image on the local filesystem. An image may have one 
    or more defined views with diferent sizes. Defined views are generated
    at first access and stored in var/ifs_views. If an image do have EXIF data
    it will be read and pickeled to var/ifs_meta for fast access."""

    meta_type = "Local Image"

    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
        )

    security = ClassSecurityInfo()  
    security.declareObjectProtected("View")    
     
    def __init__(self, id, path, uid, meta):
        self.id = id
        self.path = path
        self.title = ''
        self.convert = False
        self._meta = meta
        self._uid = uid
        self.height, self.width = self.get_dimension()

        self.content_type = "image/%s" % self._meta['IMAGE format'].lower()

        self._properties += ({'id': 'exif', 'type': 'boolean', 'mode': 'r'},)
        self.exif = self._meta.has_key('IMAGE exif')

        self.manage_options = (
            {'label':'Summary', 'action':'manage_imageSummaryForm'},
            ) + OFS.PropertyManager.PropertyManager.manage_options

        if self.exif:
            self.manage_options += ({'label':'EXIF', 'action':'manage_viewEXIF'},)
            
        self.manage_options += (
            {'label':'View', 'action':'index_html'},
            {'label':'View original', 'action':'original'},
        ) + AccessControl.Role.RoleManager.manage_options
    
    security.declarePublic('index_html', 'original', 'get_urls', 'tag')
    def index_html(self, view=None, REQUEST=None, RESPONSE=None):
        """Return the original picture."""
        t = None
        img_path = self.path
        
        if view != None:
            for i in self.aq_acquire('_thumb_views'):
                if i['id'] == view:
                    t = thumbnailer(
                            _get_cache_dirs(), 
                            (i['width'], i['height']), 
                            i['type'], 
                            i['quality'], 
                            getattr(effects, i['effect'])(),
                            getattr(framer, i['framer'])()
                            )
                    img_path, img_size = t.get(self.path, self._uid, view, make_thumb=True)
                    img_format = i['type']
                    
        if t == None:
            img_format = self._meta['IMAGE format']
            img_size = self._meta['IMAGE size']
        
        if t == None and self.aq_acquire("convert_original"):
            if self._meta['IMAGE format'] not in ["JPEG", "GIF", "PNG"]:
                t = thumbnailer(_get_cache_dirs(), 
                                ftype=self.aq_acquire("convert_to"), 
                                fquality=int(self.aq_acquire("convert_jpeg_quality")))
    
                img_file = t.convert(img_path)
                img_size = img_file.tell()
                img_file.seek(0)
                RESPONSE.setHeader('Content-Length', img_size)
                RESPONSE.setHeader('Content-Type', 'IMAGE/%s' % self.aq_acquire("convert_to"))
                
                bs = 2**16
                while True:
                    buff = img_file.read(bs)
                    RESPONSE.write(buff)
                    if len(buff) < bs: break
                return 

        RESPONSE.setHeader('Content-Length', img_size)
        RESPONSE.setHeader('Content-Type', 'IMAGE/%s' % img_format)
        return filestream_iterator(img_path, 'rb')

    def original(self, view=None, REQUEST=None, RESPONSE=None):
        """Return the original picture."""
        img_path = self.path
        img_format = self._meta['IMAGE format']
        img_size = self._meta['IMAGE size']
            
        RESPONSE.setHeader('Content-Length', img_size)
        RESPONSE.setHeader('Content-Type', 'IMAGE/%s' % img_format)
        return filestream_iterator(img_path, 'rb')

    def get_urls(self, tag=None, REQUEST=None, RESPONSE=None):
        """Return a dict with URLs to all posible views. View names as keyss."""
        tags = {}
        url = self.absolute_url()
        for i in self.aq_acquire('_thumb_views'):
            tags[i['id']] = "%s?view=%s" % (url, i['id'])
            
        return tags

    def tag(self, height=None, width=None, alt=None,
            scale=0, xscale=0, yscale=0, css_class=None, title=None, view=None, **args):
        """Return tag to a resized image."""
        if height is None: height=self.height
        if width is None:  width=self.width

        # Auto-scaling support
        xdelta = xscale or scale
        ydelta = yscale or scale

        if xdelta and width:
            width =  str(int(round(int(width) * xdelta)))
        if ydelta and height:
            height = str(int(round(int(height) * ydelta)))

        result='<img src="%s' % (self.absolute_url())
        if view != None:
            result +='?view=%s' % view
        result += '"'        

        if alt is None:
            alt=getattr(self, 'title', '')
        result = '%s alt="%s"' % (result, escape(alt, 1))

        if title is None:
            title=getattr(self, 'title', '')
        result = '%s title="%s"' % (result, escape(title, 1))

        if view == None:
            if height:
                result = '%s height="%s"' % (result, height)
    
            if width:
                result = '%s width="%s"' % (result, width)

        if not 'border' in [ x.lower() for x in  args.keys()]:
            result = '%s border="0"' % result

        if css_class is not None:
            result = '%s class="%s"' % (result, css_class)

        for key in args.keys():
            value = args.get(key)
            if value:
                result = '%s %s="%s"' % (result, key, value)

        return '%s />' % result

    security.declareProtected('Resize local image', 'size')
    def size(self, REQUEST=None):
        """Resize image on the fly to spesific size."""
        return ""

    security.declareProtected('Manage properties', 'manage_imageSummaryForm',
                                                   'manage_viewEXIF',
                                                   'get_thumb_views')
    manage_imageSummaryForm=PageTemplateFile('zpt/imageSummary.pt', globals())    
    manage_viewEXIF=PageTemplateFile('zpt/viewEXIF.pt', globals())    

    def get_thumb_views(self):
        return self.aq_acquire('_thumb_views')

    security.declareProtected('Access contents information', 'bobobase_modification_time', 
                                                             'get_size',
                                                             'get_dimension',
                                                             'get_exif',
                                                             'getContentType')
    def bobobase_modification_time(self):
        """Return when the image was last modified or the Image DateTime
        tag in EXIF if it exists."""
        if self.exif and self._meta['IMAGE exif'].has_key("Image DateTime"):
            try: return DateTime(self._meta['IMAGE exif']['Image DateTime']['value'])
            except: pass
        return DateTime(self._meta['IMAGE modified'])

    def getContentType(self):
        """Returns the content type (MIME type) of a file or image."""
        return "IMAGE/%s" % self._meta['IMAGE format']

    def get_size(self):
        """Return the size of an image in bytes"""
        return self._meta['IMAGE size']    
    
    def get_dimension(self):
        """Return the dimensions of an image as a tupple with width and height."""
        return self._meta['IMAGE dimension']    

    def get_exif(self):
        """Return an tupple with EXIF-information. Keys is EXIF-tag names. 
        Values is a tupple consisting of EXIF-tag information which is:
        field_length, field_offset, field_type, printable, tag and value."""
        if self.exif:
            return self._meta['IMAGE exif']
        else: return None

    def icon(self):
        view = self.aq_acquire('image_icon_view')
        icon = "misc_/LocalImageSystem/Image_icon.gif"    
        return self.get_urls().get(view, icon)

    def cb_isMoveable(self):
        ret = OFS.CopySupport.CopySource.cb_isMoveable(self)
        return os.access(self.path, os.R_OK | os.W_OK) and ret
        
    def cb_isCopyable(self):
        ret = OFS.CopySupport.CopySource.cb_isCopyable(self)
        return os.access(self.path, os.R_OK) and ret

    security.declareProtected('Replace local image', 'PUT')
    def PUT(self, REQUEST=None, RESPONSE=None):
        """Handle HTTP PUT requests"""
        self.dav__init(REQUEST, RESPONSE)
        
        if REQUEST.has_key("BODYFILE"):
            try:
                file(os.path.join(self.aq_parent.path, self.id), "wb").write(REQUEST["BODYFILE"].read())
            except: pass
        
        RESPONSE.setStatus(201)
        RESPONSE.setBody('')
        return RESPONSE


Globals.InitializeClass(LocalImage)


class LocalImageSystem(LocalFolder):
    """Image File System is ment to be an alternative to ExtImage. Insted
    of using a reposit directory where all image data is stored it uses the 
    structure of the filesystem. This makes it easy to publish an existing
    image library which is catalogued in the filesystem.
    
    The author's, thats me :), intension with this product is to use it as 
    a backend for an advanced web based image database run on the Zope platform.
    And as a future thought - make it Plone compatible."""


    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'wd'},
        {'id': 'path', 'type': 'string', 'mode': 'w'},
        {'id': 'convert_original', 'type': 'boolean', 'mode': 'w'},
        {'id': 'convert_jpeg_quality', 'type': 'selection', 'select_variable': 'convert_quality', 'mode': 'w'},
        {'id': 'image_icon_view', 'type': 'selection', 'select_variable': 'image_icon', 'mode': 'w'},
        )

    manage_options = LocalFolder.manage_options[:2] + \
                     ({'label':'Image views', 'action':'manage_thumbnailViewsForm'},) + \
                     ({'label':'Image cache', 'action':'manage_cacheForm'},) + \
                     LocalFolder.manage_options[2:]
    
    meta_type = "Image File System"
    isPrincipiaFolderish = 1
    
    security = ClassSecurityInfo()
    
    def __init__(self, id, title, path):
        LocalFolder.__init__(self, id, path)
        self.title = title
        #~ self.root_id = id # the id of the original product
        self.convert_original = False
        self.convert_to = 'JPEG'
        self.convert_quality = [str(i) for i in range(99,0,-1)]
        self.convert_jpeg_quality = '75'
        self.resize_to = 'JPEG'
        self.resize_jpeg_quality ='75'
        self.resize_quality = [str(i) for i in range(99,0, -1)]
        self.image_icon_view = 'icon'
        
        self._thumb_views = [ 
            {"id": "icon", "width": 32, "height": 24, "type": "JPEG", "quality": 75, "framer": "ThinFrame", "effect": "Effect"},
            {"id": "thumbnail", "width": 100, "height": 100, "type": "JPEG", "quality": 75, "framer": "ThinShadowFrame", "effect": "Effect"}
            ]

    security.declareProtected('Manage properties', 'manage_thumbnailViewsForm',
                                                   'manage_cacheForm',
                                                   'manage_deleteThumbnailView',
                                                   'manage_addThumbnailView',
                                                   'manage_clearImageCache',
                                                   'manage_clearMetaCache',
                                                   'meta_cache_size',
                                                   'image_cache_size',
                                                   'get_views',
                                                   'get_framers',
                                                   'get_effects')

    def manage_clearImageCache(self, REQUEST=None):
        """Clear image disk image cache."""
        
        t = thumbnailer(_get_cache_dirs())        
        t.flush()
        
        if REQUEST != None:
            return self.manage_cacheForm(REQUEST, 
                management_view = "Image cache",
                manage_tabs_message = "Image cache cleared."
                )

    def manage_clearMetaCache(self, REQUEST=None):
        """Clear image disk image cache."""
        
        self.clear_meta(self)
        
        if REQUEST != None:
            return self.manage_cacheForm(REQUEST, 
                management_view = "Image cache",
                manage_tabs_message = "Meta cache cleared."
                )
                
    def clear_meta(self, parent):
        for i in parent.objectValues():
            if i.meta_type == "Local Folder":
                self.clear_meta(i)

    def manage_addThumbnailView(self, view, REQUEST=None):
        """Add a thumbnail view."""
        self._thumb_views.append(view)
        self._p_changed = 1
        
        if REQUEST != None:
            return self.manage_thumbnailViewsForm(REQUEST, 
                management_view = "Image views",
                manage_tabs_message = "Thumbnail view added."
                )
        
        
    def manage_deleteThumbnailView(self, id, REQUEST=None):
        """Remove a thumbnail view."""
            
        hits = []    
        for j in range(len(self._thumb_views)):
            for i in id:
                if i == self._thumb_views[j]['id']: 
                    hits.append(j)
                    break
        
        hits.reverse()        
        
        for i in hits: del self._thumb_views[i]
        self._p_changed = 1

        if hits == []: ret = "No view selected for delete."
        else: ret = "Thumbnail view(s) deleted."

        if REQUEST != None:
            return self.manage_thumbnailViewsForm(REQUEST, 
                management_view = "Image views",
                manage_tabs_message = ret
                )
    
    manage_thumbnailViewsForm=PageTemplateFile('zpt/thumbnailViews.pt', globals())    
    manage_cacheForm=PageTemplateFile('zpt/cache.pt', globals())    
    
    def get_views(self):
        """Return supported thumbnail views."""
        
        return self._thumb_views

    def get_framers(self):
        """Return description and class name of available framers as 
        list of tuples."""
        return self._get_file_classes(framer, framer.Frame)        

    def get_effects(self):
        """Return description and class name of available effects as 
        list of tuples."""
        return self._get_file_classes(effects, effects.Effect)        

    def meta_cache_size(self):
        return 0
        
    def image_cache_size(self):
        size = 0
        count = 0
        for i in _get_cache_dirs():
            for root, dirs, files in os.walk(i):
                for j in files:
                    s = os.stat(os.path.join(root, j))
                    
                    count += 1
                    size += s[stat.ST_SIZE]
                break
              
        return count, size
                
    def _get_file_classes(self, pyfile, subclassof):
        c = dir(pyfile)
        classes = []
        for i in c:
            if callable(getattr(pyfile, i)) and issubclass(getattr(pyfile, i), subclassof):
                classes.append(
                    {"id": i, "name": str(getattr(pyfile, i)())}
                    )
        return classes
        
    def image_icon(self):
        ret = ['None']
        for i in self.get_views():
            ret.append(i['id'])
        return ret

Globals.InitializeClass(LocalImageSystem)

def manage_addLIS(dispatcher, id, title, path, REQUEST=None):
    """
    Create a new Image File System object and adds it in zope.
    """
    dispatcher.Destination()._setObject(id, LocalImageSystem(id, title, path))
    if REQUEST != None:
        return dispatcher.manage_main(dispatcher, REQUEST) 

manage_addLISForm=Globals.DTMLFile('dtml/LISAdd', globals())
